from ..repository import BaseRepository
from ..schema import ListFilter,FindUniqueByFieldInput,FilterSchema,EmbeddingSchema
from typing import Any, Dict, Optional,List

class BaseService:
    def __init__(self, repository: BaseRepository):
        self.repository = repository

    async def create(self, data: dict, collection_name: str = None, embedding_config: EmbeddingSchema = None) -> any:
        return await self.repository.create(data, collection_name, embedding_config)
    
    async def bulk_create(self, data: list[any], collection_name: str = None, embedding_config: EmbeddingSchema = None) -> list[any]:
        return await self.repository.bulk_create(data, collection_name, embedding_config)
    
    async def get_by_attr(self, attr: str, value: any, collection_name: str = None) -> any:
        return await self.repository.get_by_attr(attr, value, collection_name)

    async def update(self, _id: str, data: dict, collection_name: str = None, embedding_config: EmbeddingSchema = None) -> any:
        return await self.repository.update(_id, data, collection_name, embedding_config)
    
    async def delete(self, _id: str, collection_name: str = None, embedding_config: EmbeddingSchema = None) -> any:
        return await self.repository.delete(_id, collection_name, embedding_config)

    async def get_unique_values(self, schema: FindUniqueByFieldInput, collection_name: str = None) -> list[any]:
        return await self.repository.get_unique_values(schema, collection_name)
    
    async def get_all(self,find:ListFilter=dict,collection_name:str=None)->list[any]:
        return await self.repository.get_all(find,collection_name)
    
    async def bulk_update(self, conditions: FilterSchema, update_data: Dict[str, Any], collection_name: Optional[str] = None, embedding_config: EmbeddingSchema = None) -> int:
        return await self.repository.bulk_update(conditions, update_data, collection_name, embedding_config)
    
    async def bulk_delete(self, conditions: FilterSchema, collection_name: Optional[str] = None, embedding_config: EmbeddingSchema = None) -> int:
        return await self.repository.bulk_delete(conditions, collection_name, embedding_config)
    
    async def bulk_write(self, data: List[any], collection_name: Optional[str] = None):
        return await self.repository.bulk_write(data, collection_name)


from datetime import datetime
from functools import lru_cache
from typing import Any, Union, Dict, List
from bson import ObjectId
# Cache for datetime format parsing
@lru_cache(maxsize=128)
def parse_datetime(value: str) -> Union[datetime, str]:
    """
    Parse datetime string with caching for better performance.
    Supports multiple ISO 8601 formats.
    """
    formats = [
        "%Y-%m-%dT%H:%M:%S",       # e.g., 2025-05-10T14:00:00
        "%Y-%m-%dT%H:%M:%SZ",      # e.g., 2025-05-10T14:00:00Z
        "%Y-%m-%dT%H:%M:%S.%f",    # e.g., 2025-05-10T14:00:00.123456
        "%Y-%m-%dT%H:%M:%S.%fZ",   # e.g., 2025-05-10T14:00:00.123456Z
        "%Y-%m-%d"                 # e.g., 2025-05-10
    ]
    
    for fmt in formats:
        try:
            return datetime.strptime(value, fmt)
        except ValueError:
            continue
    return value

# Cache for boolean string values
BOOLEAN_MAP = {
    "true": True, 
    "false": False
}

def coerce_value(value: Any) -> Any:
    """
    Coerce a value to its appropriate Python type.
    Handles nested structures (lists and dicts) recursively.
    Optimized for common cases and includes caching for better performance.
    """
    # Fast path for non-string types
    if not isinstance(value, str):
        if isinstance(value, list):
            return [coerce_value(item) for item in value]
        if isinstance(value, dict):
            return {k: coerce_value(v) for k, v in value.items()}
        return value

    # Fast path for empty string
    if not value:
        return value

    # Check boolean/null values first (most common case)
    lowered = value.lower()
    if lowered in BOOLEAN_MAP:
        return BOOLEAN_MAP[lowered]
    
    try:
        if ObjectId.is_valid(value):
            return ObjectId(value)
    except Exception as e:
        pass
    
    # Check for datetime (least common case)
    return parse_datetime(value)



from beanie import Document, PydanticObjectId
from pydantic import Field
from uuid import uuid4
from datetime import datetime,UTC
class BaseDocument(Document):
    """
    Base document class for all documents of NoSQL database
    """
    id: PydanticObjectId = Field(default_factory=PydanticObjectId, unique=True, index=True,nullable=False,alias="_id")
    uuid: str = Field(default_factory=lambda: str(uuid4()), unique=True, index=True,nullable=False)
    created_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
    updated_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
        
class BaseDraftDocument(BaseDocument):
    is_draft: bool = Field(default=False)

        


from pydantic import BaseModel, ConfigDict, Field
from datetime import datetime
from enum import Enum
from typing import Optional, Union, List, Any, Literal

class Operator(str, Enum):
    EQ = "eq"
    NE = "ne"
    GT = "gt"
    GTE = "gte"
    LT = "lt"
    LTE = "lte"
    IN = "in"
    NIN = "nin"
    LIKE = "like"
    ILIKE = "ilike"
    BETWEEN = "between"
    IS_NULL = "is_null"
    IS_NOT_NULL = "is_not_null"
    MIN_LENGTH = "min_length"
    MAX_LENGTH = "max_length"

class LogicalOperator(str, Enum):
    AND = "and"
    OR = "or"

class SortDirection(str, Enum):
    ASC = "asc"
    DESC = "desc"

class FieldTypeEnum(str, Enum):
    STR = "str"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    LIST = "list"
    DICT = "dict"
    DATETIME = "datetime"

class BaseSchema(BaseModel):
    _id: str
    uuid: str
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(from_attributes=True)

class BaseDraftSchema(BaseSchema):
    is_draft: bool = Field(default=True)

# Schema for getting distinct values of a field
class FindUniqueByFieldInput(BaseModel):
    field_name: str
    ordering: Optional[Literal["asc", "desc"]] = None
    page: Optional[int] = Field(default=1)
    page_size: Optional[int] = Field(default=10)
    search: Optional[str] = None

# Primitive field condition
class FieldOperatorCondition(BaseModel):
    field: str
    operator: Operator
    value: Any

# Base structure for a logical group
class LogicalCondition(BaseModel):
    operator: LogicalOperator
    conditions: List["ConditionType"]

# Each item in conditions list can be:
# 1. a logical condition (nested group)
# 2. a dict like {field: ..., operator: ..., value: ...}
ConditionType = Union["LogicalCondition", "FieldOperatorCondition"]

# Top-level filter schema
class FilterSchema(BaseModel):
    operator: LogicalOperator
    conditions: List[ConditionType]

# Sort schema
class SortSchema(BaseModel):
    field: str
    direction: SortDirection

# Schema for displaying search operations
class SearchOptions(BaseModel):
    search: Optional[str] = None
    sort_order: Optional[List[SortSchema]] = None
    page: Optional[int] = None
    page_size: Optional[int] = None
    total_pages: Optional[int] = None
    total_count: Optional[int] = None

# Schema for displaying find operations' result
class ListResponse(BaseModel):
    founds: List[Any]
    search_options: SearchOptions

class AggregationType(str, Enum):
    SUM = "sum"
    AVG = "avg"
    COUNT = "count"
    MAX = "max"
    MIN = "min"

class AggregationSpec(BaseModel):
    field: str
    type: AggregationType
    alias: str

class GroupOptions(BaseModel):
    page: Optional[int] = Field(default=1, description="Page number for group pagination")
    page_size: Optional[int] = Field(default=10, description="Number of groups per page")
    group_id: Optional[dict] = Field(default=None, description="Filter to apply to grouped _id after grouping (was group_filter)")
    sort_by: Optional[List[SortSchema]] = Field(default=None, description="Sort order for groups (columns) (was group_sort_order)")

class Aggregations(BaseModel):
    group_by: List[str]
    document_inclusion_mode: Optional[Literal["none", "all", "partial"]] = Field(
        default="none",
        description="Options: 'none', 'all', 'partial' . 'none' means no document fields will be included in the response, 'all' means all document fields will be included in the response, 'partial' means only the document fields that match the included_fields will be included in the response"
    )
    included_fields: Optional[List[str|dict]] = Field(
        default=None,
        description="Fields to include when document_inclusion_mode is 'partial'"
    )
    aggregation_fields: Optional[List[AggregationSpec]] = None
    group_options: Optional[GroupOptions] = Field(default=None, description="Group-level pagination, filtering, and sorting options")

class FieldRule(BaseModel):
    action: Literal["include", "exclude"]
    field: str
    conditions: Optional[FilterSchema] = None

class GraphLookup(BaseModel):
    max_depth: Optional[int] = Field(default=1)
    connect_from_field: str = Field(default=None)
    connect_to_field: str = Field(default=None)

class ReferenceField(BaseModel):
    local_field: str = Field(description="The field of the local entity")
    foreign_field: str = Field(description="The field of the foreign entity")
    foreign_entity: str = Field(description="The name of the foreign entity")
    alias: str = Field(description="The alias of the foreign entity")
    view_fields: Optional[List[str]] = None
    use_graph_lookup: Optional[GraphLookup] = Field(default=None)
    sort_fields: Optional[List[SortSchema]] = Field(default=None, description="Fields to sort the referenced documents by")
    single_object: Optional[bool] = Field(default=False, description="If true, the referenced document will be returned as a single object")

# Type for aggregation pipeline stages
class AggregationStage(BaseModel):
    pipeline: list = Field(description="The pipeline configuration for the stage")

# Union type for reference fields that can be either ReferenceField or AggregationStage
ReferenceFieldType = Union[List[ReferenceField], AggregationStage]

# Field filter schema for including/excluding specific fields
class FieldFilterSchema(BaseModel):
    type: Literal["include", "exclude"] = Field(
        description="Whether to include or exclude the specified fields"
    )
    fields: List[str|dict] = Field(
        description="List of field names to include or exclude"
    )
class ListFilter(BaseModel):
    pre_filters: Optional[FilterSchema] = None
    filters: Optional[FilterSchema] = None
    sort_order: Optional[List[SortSchema]] = None
    page: Optional[int] = Field(default=1)
    page_size: Optional[int] = Field(default=20)
    search: Optional[str] = None
    searchable_fields: Optional[List[str]] = None
    field_rules: Optional[List[FieldRule]] = None
    reference_fields: Optional[ReferenceFieldType] = None
    aggregations: Optional[Aggregations] = None
    field_filter: Optional[FieldFilterSchema] = None

class EmbeddingSchema(BaseModel):
    container_name: str


from typing import Any
from datetime import datetime
from decimal import Decimal
from bson import ObjectId

def convert_to_text(data: Any) -> str:
    if data is None:
        return ""

    if isinstance(data, (int, float, bool, datetime, Decimal, ObjectId)):
        return str(data)
    
    if isinstance(data, (list, tuple, set)):
        return ", ".join(convert_to_text(item) for item in data)
    
    if isinstance(data, dict):
        return ", ".join(f"{key}: {convert_to_text(value)}" for key, value in data.items())
    
    if isinstance(data, str):
        return data
    
    return str(data)


from operator import index
from typing import Type, TypeVar, List, Dict, Any, Union, Optional
from abs_exception_core.exceptions import (
    BadRequestError,
    NotFoundError,
    GenericHttpError,
    ValidationError,
)
from pydantic import BaseModel
from pymongo import ASCENDING, DESCENDING
from beanie.operators import Set
from pymongo.errors import PyMongoError
from datetime import datetime, UTC
from motor.motor_asyncio import AsyncIOMotorDatabase
from beanie import Document
from bson import ObjectId, Decimal128
from decimal import Decimal
from uuid import uuid4
from abs_utils.azure_service_bus.azure_service_bus import AzureServiceBus
from abs_utils.socket_io.server import SocketIOService
from abs_utils.constants.event_constants import (
    ON_RECORD_CREATION,
    ON_RECORD_UPDATION,
    ON_RECORD_DELETION,
)
from ..schema import (
    ListFilter,
    SortDirection,
    Operator,
    FindUniqueByFieldInput,
    FieldRule,
    Aggregations,
    AggregationType,
    ReferenceField,
    FilterSchema,
    SortSchema,
    ReferenceFieldType,
    AggregationStage,
    EmbeddingSchema
)
from ..util import coerce_value, apply_condition, logical_operator_map, convert_to_text
from abs_utils.logger import setup_logger

logger = setup_logger(__name__)
T = TypeVar("T", bound=BaseModel)
DocumentType = TypeVar("DocumentType", bound=Document)


class BaseRepository:
    """
    Base repository class for doing all the database operations using Beanie for NoSQL database.
    """

    def __init__(
        self,
        document: Type[DocumentType] = None,
        db: AsyncIOMotorDatabase = None,
        sio_service: SocketIOService = None,
        azure_service_bus: AzureServiceBus = None,
    ):
        if document is None and db is None:
            raise ValidationError(detail="Either document or db must be provided")
        self.document = document
        self.db = db
        self.sio_service = sio_service
        self.azure_service_bus = azure_service_bus

    def _convert_to_json_serializable(self, data: Any) -> Any:
        """
        Converts MongoDB documents to JSON-serializable format.
        """
        if data is None:
            return None

        # Fast path for common types
        if isinstance(data, (str, int, float, bool)):
            return data

        # Handle MongoDB specific types
        if isinstance(data, (ObjectId, Decimal128)):
            return str(data)

        if isinstance(data, datetime):
            return data.isoformat()

        if isinstance(data, Decimal):
            return float(data)

        # Handle collections
        if isinstance(data, dict):
            return {
                str(k): self._convert_to_json_serializable(v) for k, v in data.items()
            }

        if isinstance(data, (list, tuple, set)):
            return [self._convert_to_json_serializable(item) for item in data]

        # Handle Pydantic and Beanie models
        if isinstance(data, (BaseModel, Document)):
            return self._convert_to_json_serializable(data.model_dump())

        # Handle objects with __dict__
        if hasattr(data, "__dict__"):
            return self._convert_to_json_serializable(vars(data))

        # If we can't convert it, return string representation
        return str(data)

    def get_collection(self, collection_name: Optional[str] = None) -> Any:
        """Get the collection from the database"""
        return (
            self.db.get_collection(collection_name)
            if collection_name
            else self.document.get_motor_collection()
        )

    def get_base_document_fields(self) -> Dict[str, Any]:
        """Get the base document fields"""
        return {
            "uuid": str(uuid4()),
            "created_at": datetime.now(UTC),
            "updated_at": datetime.now(UTC),
        }

    async def _handle_mongo_error(self, operation: str, error: Exception) -> None:
        """Handle MongoDB errors consistently."""
        if isinstance(error, PyMongoError):
            raise GenericHttpError(
                status_code=500,
                detail=str(error),
                error_type="PyMongoError",
                message=f"Failed to {operation}",
            )
        raise error

    def _coerce_document(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Coerce all values in a document to their appropriate types.
        This is used before saving to the database to ensure proper type conversion.
        """
        return coerce_value(data)

    async def create(
        self, obj: T, collection_name: Optional[str] = None, embedding_config: Optional[EmbeddingSchema] = None
    ) -> Dict[str, Any]:
        """Create a new document in the collection"""
        try:
            # Convert to dict and coerce values
            obj_dict = obj.model_dump() if hasattr(obj, "model_dump") else dict(obj)
            obj_dict = self._coerce_document(obj_dict)

            if collection_name:
                collection = self.get_collection(collection_name)
                result = await collection.insert_one(
                    {**self.get_base_document_fields(), **obj_dict}
                )
                created_doc = await self.get_by_attr(
                    "_id", result.inserted_id, collection_name
                )
                entity_id = collection_name.split("_")[1]
                try:
                    if embedding_config and self.azure_service_bus:
                        payload = {
                            "event_id": str(uuid4()),
                            "event_key":"embedding",
                            "action":"creation",
                            "container_name": embedding_config.container_name,
                            "payload": [
                                {
                                    "id": str(result.inserted_id),
                                    "text": convert_to_text(created_doc),
                                    "metadata": {
                                        "entity_id": entity_id,
                                        "content":{
                                            "id":str(created_doc["_id"]),
                                            "created_at":created_doc["created_at"],
                                            "updated_at":created_doc["updated_at"],
                                            "uuid":created_doc["uuid"],
                                            "data":obj_dict
                                        }
                                    }
                                }
                            ]
                        }

                        await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                except Exception as e:
                    logger.error(f"Error sending embedding payload: {e}")
                    
                if self.sio_service or self.azure_service_bus:
                    payload = {
                        "event_id": str(uuid4()),
                        "event_key": ON_RECORD_CREATION,
                        "result": created_doc,
                        "payload": obj_dict,
                        "record_id": str(result.inserted_id),
                        "collection_name": collection_name,
                    }
                    if self.sio_service:
                        await self.sio_service.emit_to_room(
                            ON_RECORD_CREATION,f"entity:{entity_id}" ,payload
                        )
                    if self.azure_service_bus:
                        await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                return created_doc

            model_instance = self.document(**obj_dict)
            model = await model_instance.insert()
            return await self.get_by_attr("id", model.id)

        except Exception as e:
            await self._handle_mongo_error("create document", e)

    async def bulk_create(
        self, data: List[T], collection_name: Optional[str] = None, embedding_config: Optional[EmbeddingSchema] = None
    ) -> List[Dict[str, Any]]:
        """Create multiple documents in the collection"""
        try:
            # Convert to dicts and coerce values
            get_obj = lambda obj: (
                obj.model_dump() if hasattr(obj, "model_dump") else dict(obj)
            )
            coerced_data = [self._coerce_document(get_obj(item)) for item in data]

            if collection_name:
                collection = self.get_collection(collection_name)
                documents = [
                    {**self.get_base_document_fields(), **item} for item in coerced_data
                ]
                result = await collection.insert_many(documents)
                result_docs = []
                for doc in documents:
                    doc["_id"] = str(doc.get("_id", ""))
                    result_docs.append(self._convert_to_json_serializable(doc))
                # Convert the documents to JSON serializable format
                created_docs = []
                entity_id = collection_name.split("_")[1]
                try:
                    if embedding_config and self.azure_service_bus:
                        for doc in result_docs:
                            doc_data = {}
                            doc_data["id"] = str(doc.get("_id", ""))
                            doc_data["text"] = convert_to_text(doc)
                            doc_data["metadata"] = {
                                "entity_id": entity_id,
                                "content":{
                                    "id":str(doc.get("_id", "")),
                                    "created_at":doc.get("created_at", ""),
                                    "updated_at":doc.get("updated_at", ""),
                                    "uuid":doc.get("uuid", ""),
                                    "data":doc
                                }
                            }
                            created_docs.append(doc_data)

                        payload = {
                            "event_id": str(uuid4()),
                            "event_key": "embedding",
                            "action": "creation",
                            "container_name": embedding_config.container_name,
                            "payload": created_docs
                        }

                        await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                except Exception as e:
                    logger.error(f"Error sending embedding payload: {e}")
                    
                # TODO: Add event handling for bulk create
                # payload = {
                #     "event_id": str(uuid4()),
                #     "event_key": "records_created",
                #     "payload": created_docs,
                #     "request_body": coerced_data,
                #     "collection_name": collection_name,
                # }
                # if self.sio_service:
                #     await self.sio_service.emit_broadcast(
                #         "records_created", payload
                #     )
                # if self.azure_service_bus:
                #     await self.azure_service_bus.send(payload)
                return result_docs

            model_instances = [self.document(**item) for item in coerced_data]
            await self.document.insert_many(model_instances)
            return [
                self._convert_to_json_serializable(doc.model_dump())
                for doc in model_instances
            ]

        except Exception as e:
            await self._handle_mongo_error("bulk create documents", e)

    async def bulk_write(self, data: List[T], collection_name: Optional[str] = None):
        """Bulk write documents to the collection"""
        try:  
            collection = self.get_collection(collection_name)
            result = await collection.bulk_write(data,ordered=False)
            return result.upserted_count
        except Exception as e:
            await self._handle_mongo_error("bulk write documents", e)

    async def update(
        self, id: Union[str, ObjectId], obj: T, collection_name: Optional[str] = None, embedding_config: Optional[EmbeddingSchema] = None
    ) -> Dict[str, Any]:
        """Update a document by id"""
        try:
            # remove none values from the object
            if hasattr(obj, "model_dump"):
                obj_dict = obj.model_dump(exclude_none=True)
            else:
                obj_dict = {k: v for k, v in obj.items() if v is not None}
            # Convert to dict and coerce values
            obj_dict = self._coerce_document(obj_dict)
            object_id = ObjectId(id) if isinstance(id, str) else id

            if collection_name:
                collection = self.get_collection(collection_name)
                result = await collection.update_one(
                    {"_id": object_id},
                    {"$set": obj_dict, "$currentDate": {"updated_at": True}},
                )

                if result.matched_count == 0:
                    raise NotFoundError(detail=f"Document with id {id} not found")

                updated_doc = await self.get_by_attr("_id", object_id, collection_name)
                entity_id = collection_name.split("_")[1]
                try:
                    if embedding_config and self.azure_service_bus:
                        payload = {
                            "event_id": str(uuid4()),
                            "event_key": "embedding",
                            "action": "updation",
                            "container_name": embedding_config.container_name,
                            "payload": [
                                {
                                    "id": str(object_id),
                                    "text": convert_to_text(updated_doc),
                                    "metadata": {
                                        "entity_id": entity_id,
                                        "content":{
                                            "id":str(updated_doc.get("_id", "")),
                                            "created_at":updated_doc.get("created_at", ""),
                                            "updated_at":updated_doc.get("updated_at", ""),
                                            "uuid":updated_doc.get("uuid", ""),
                                            "data":updated_doc
                                        }
                                    }
                                }
                            ]
                        }

                        await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                except Exception as e:
                    logger.error(f"Error sending embedding payload: {e}")

                if self.sio_service or self.azure_service_bus:
                    payload = {
                        "event_id": str(uuid4()),
                        "event_key": ON_RECORD_UPDATION,
                        "result": updated_doc,
                        "payload": obj_dict,
                        "record_id": str(object_id),
                        "collection_name": collection_name,
                    }
                    if self.sio_service:
                        await self.sio_service.emit_to_room(
                            ON_RECORD_UPDATION,f"entity:{entity_id}" ,payload
                        )
                    if self.azure_service_bus:
                        await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                return updated_doc

            obj_dict["updated_at"] = datetime.now(UTC)
            result = await self.document.get(object_id)
            if not result:
                raise NotFoundError(detail=f"Document with id {id} not found")

            await result.update(Set(obj_dict))
            return await self.get_by_attr("id", object_id)

        except Exception as e:
            await self._handle_mongo_error("update document", e)

    async def get_by_attr(
        self,
        attr: Union[str, Dict[str, Any]],
        value: Any = None,
        collection_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        try:
            # Handle multiple attributes
            if isinstance(attr, dict):
                query = {}
                for field, field_value in attr.items():
                    if field == "id":
                        field = "_id"
                        field_value = ObjectId(field_value)
                    query[field] = field_value
            else:
                if attr == "id":
                    attr = "_id"
                    value = ObjectId(value)
                query = {attr: value}

            if collection_name:
                collection = self.get_collection(collection_name)
                result = await collection.find_one(query)
                if not result:
                    raise NotFoundError(
                        detail=f"Document with {attr}={value} not found"
                    )
                return self._convert_to_json_serializable(result)

            if not self.document:
                raise ValueError("Document class is not provided.")

            result = await self.document.find_one(query)
            if not result:
                raise NotFoundError(detail=f"Document with {attr}={value} not found")
            return self._convert_to_json_serializable(result)

        except Exception as e:
            await self._handle_mongo_error("get document", e)

    async def get_many_by_attr(
        self,
        attr: Union[str, Dict[str, Any]],
        value: Any = None,
        collection_name: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """Get multiple documents by attribute"""
        try:
            if isinstance(attr, dict):
                query = {}
                for field, field_value in attr.items():
                    if field == "id":
                        field = "_id"
                        field_value = ObjectId(field_value)
                    query[field] = field_value
            else:
                if attr == "id":
                    attr = "_id"
                    value = ObjectId(value)
                query = {attr: value}

            if collection_name:
                collection = self.get_collection(collection_name)
                result = await collection.find(query).to_list(length=None)
                return [self._convert_to_json_serializable(doc) for doc in result]

            if not self.document:
                raise ValueError("Document class is not provided.")

            result = await self.document.find(query).to_list(length=None)
            return [self._convert_to_json_serializable(doc) for doc in result]
        except Exception as e:
            await self._handle_mongo_error("get many documents", e)

    
    def _build_search_conditions(
        self, search_term: str, searchable_fields: List[str]
    ) -> List[Dict[str, Any]]:
        """Build case-insensitive search conditions for fields"""
        regex = {"$regex": f".*{search_term}.*", "$options": "i"}
        return [{field: regex} for field in searchable_fields]
    

    def _build_query_filter(
        self, find: ListFilter, collection_name: Optional[str] = None
    ) -> Optional[Dict[str, Any]]:
        """Build MongoDB filter from ListFilter"""
        if find.filters:
            if hasattr(find.filters, "model_dump"):
                filter = find.filters.model_dump(exclude_none=True)
            else:
                filter = {k: v for k, v in find.filters.items() if v is not None}
        else:
            filter = {}

        base_filter = (
            self._build_filter_condition(filter, collection_name, is_expr=False)
            if find.filters
            else None
        )
        if find.search and find.searchable_fields:
            search_filter = {
                "$or": self._build_search_conditions(
                    find.search, find.searchable_fields
                )
            }
            return (
                {"$and": [base_filter, search_filter]} if base_filter else search_filter
            )

        return base_filter


    def _build_filter_condition(
        self,
        filter_dict: Dict[str, Any],
        collection_name: Optional[str] = None,
        is_expr: bool = False,
    ) -> Dict[str, Any]:
        """Recursively build MongoDB filter"""
        if not isinstance(filter_dict, dict):
            return {}

        if "operator" in filter_dict and "conditions" in filter_dict:
            conditions = [
                self._build_filter_condition(cond, collection_name, is_expr)
                for cond in filter_dict["conditions"]
            ]
            return {logical_operator_map[filter_dict["operator"]]: conditions}

        if all(k in filter_dict for k in ["field", "operator", "value"]):
            try:
                model = self.document or self.db[collection_name]
                op = Operator(filter_dict["operator"].lower())
                if filter_dict["field"] == "id" or filter_dict["field"] == "_id":
                    if isinstance(filter_dict["value"], list):
                        value = [ObjectId(v) for v in filter_dict["value"]]
                    else:
                        value = ObjectId(filter_dict["value"])
                else:
                    value = coerce_value(filter_dict["value"])
                return apply_condition(model, op, filter_dict["field"], value, is_expr)
            except ValueError:
                raise BadRequestError(
                    f"Invalid comparison operator: {filter_dict['operator']}"
                )

        return {}

    def _get_sort_order(self, sort_order: List[SortSchema]) -> List[tuple]:
        """Generate sort tuples for MongoDB"""
        return [
            (s.field, ASCENDING if s.direction == SortDirection.ASC else DESCENDING)
            for s in sort_order
        ] or [("_id", DESCENDING)]
    
    async def get_all(
        self, find: ListFilter, collection_name: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """Retrieve documents with filtering, sorting, and pagination"""
        try :
            page = find.page or 1
            page_size = find.page_size or 20
            skip = (page - 1) * page_size
            pipeline = []

            if find.pre_filters:
                if hasattr(find.pre_filters, "model_dump"):
                    pre_filters = find.pre_filters.model_dump(exclude_none=True)
                else:
                    pre_filters = {
                        k: v for k, v in find.pre_filters.items() if v is not None
                    }
                pipeline.append(
                    {
                        "$match": self._build_filter_condition(
                            pre_filters, collection_name, is_expr=False
                        )
                    }
                )

            if find.reference_fields:
                pipeline = self._build_reference_stage(pipeline, find.reference_fields)

            query_filter = self._build_query_filter(find, collection_name)
            if query_filter:
                pipeline.append({"$match": query_filter})

            if find.field_rules:
                pipeline = self._get_field_rule(find.field_rules, pipeline, collection_name)

            # Handle field_filter for non-aggregation queries
            if find.field_filter and not find.aggregations:
                field_filters = self._build_field_filters(find.field_filter)
                if field_filters:
                    pipeline.append({"$project": field_filters})

            if find.sort_order:
                pipeline.append({"$sort": self._get_sort_order(find.sort_order)})

            model = self.document or self.db[collection_name]

            count_pipeline = pipeline + [{"$count": "total"}]
            count_result = await model.aggregate(count_pipeline).to_list(length=1)
            total_count = count_result[0]["total"] if count_result else 0

            if find.aggregations:
                # After: Single method call
                projection = self._build_aggregation_projection(find) if find.aggregations.group_by else None
                if projection:
                    pipeline.append({"$project": projection})


                group_stage = self._build_group_stage(find.aggregations,find.field_filter)
                if "$group" in group_stage and "count" not in group_stage["$group"]:
                    group_stage["$group"]["count"] = {"$sum": 1}

                pipeline.append(group_stage)

                
                # Apply group filtering (for group_id filtering, not field filtering)
                pipeline += self._build_group_filter_stage(find.aggregations)


                pipeline.append(self._build_data_pagination_stage(find.aggregations))


                data = await model.aggregate(pipeline).to_list(length=None)

                count_pipeline = self._build_count_pipeline(pipeline)
                count_result = await model.aggregate(count_pipeline).to_list(length=1)
                total_groups = count_result[0]["total"] if count_result else len(data)
                group_options = getattr(find.aggregations, 'group_options', None)
                group_page = getattr(group_options, 'page', 1) if group_options else 1
                group_page_size = getattr(group_options, 'page_size', 10) if group_options else 10
                total_group_pages = (total_groups + group_page_size - 1) // group_page_size

                data, data_page, data_page_size = self._process_group_data(data, find)

                result = {
                    "founds": self._convert_to_json_serializable(data),
                    "total_groups": total_groups,
                    "search_options": {
                        "total_pages": None,
                        "total_count": total_count,
                        "page": None,
                        "page_size": None,
                        "search": find.search,
                        "sort_order": find.sort_order,
                    },
                }
                return result
            else:
                pipeline += [{"$skip": skip}, {"$limit": page_size}]
                data = await model.aggregate(pipeline).to_list(length=None)
                total_pages = (total_count + page_size - 1) // page_size

                return {
                    "founds": self._convert_to_json_serializable(data),
                    "search_options": {
                        "total_pages": total_pages,
                        "total_count": total_count,
                        "page": page,
                        "page_size": page_size,
                        "search": find.search,
                        "sort_order": find.sort_order,
                    },
                }
        except Exception as e:
            await self._handle_mongo_error("get_all", e)

    def _build_field_rules_stage(
        self, field_rules: List[FieldRule], collection_name: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """Build $addFields stages for dynamic array inclusion/exclusion logic"""
        stages = []

        for rule in field_rules:
            action = rule.action.lower()
            field_path = rule.field
            conditions = rule.conditions

            if action not in ("include", "exclude"):
                raise ValueError(f"Unsupported filter action: {action}")
            
            if not conditions or not conditions.conditions:
                continue

            condition_expr = self._build_filter_condition(
                filter_dict=conditions.model_dump(),
                collection_name=collection_name,
                is_expr=True,
            )

            filter_condition = ( {"$not": [condition_expr]} if action == "exclude" else condition_expr)

            stages.append(
                {
                    "$addFields": {
                        field_path: {
                            "$filter": {
                                "input": f"${field_path}",
                                "as": "item",
                                "cond": filter_condition,
                            }
                        }
                    }
                }
            )

        return stages
    
    def _get_field_rule(
        self,
        field_rules: List[FieldRule],
        pipeline: List[Dict[str, Any]],
        collection_name: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """Append field modifier stages to the aggregation pipeline."""
        if not field_rules:
            return pipeline

        field_rule_stages = self._build_field_rules_stage(field_rules, collection_name)
        pipeline.extend(field_rule_stages)
        return pipeline

    

    def _build_reference_stage(
        self, pipeline: List[Dict[str, Any]], references: ReferenceFieldType
    ) -> List[Dict[str, Any]]:
        """Build reference stages supporting both flat and deeply nested alias paths."""
        if isinstance(references,AggregationStage):
            return pipeline+references.pipeline
        else:
            def process_reference(ref, current_pipeline):
                local_path = ref.local_field
                foreign_field = ref.foreign_field or "_id"
                from_collection = ref.foreign_entity
                alias = ref.alias or local_path.split(".")[-1] + "_info"
                view_fields = ref.view_fields or []
                is_single = ref.single_object or False
                sort_fields = ref.sort_fields or []

                # Convert sort_fields to MongoDB sort dict
                ref_sort_dict = {}
                for sort_item in sort_fields:
                    direction = 1 if sort_item.direction == SortDirection.ASC else -1
                    ref_sort_dict[sort_item.field] = direction

                path_parts = local_path.split(".")
                is_nested_local = len(path_parts) > 1
                alias_parts = alias.split(".")
                is_nested_alias = len(alias_parts) > 1

                local_key = path_parts[-1]
                embed_field = alias_parts[-1]

                # Prepare lookup pipeline
                lookup_pipeline = [
                    {
                        "$match": {
                            "$expr": {
                                "$in": [
                                    f"${foreign_field}",
                                    {
                                        "$cond": {
                                            "if": {"$eq": [{"$type": "$$ids"}, "array"]},
                                            "then": "$$ids",
                                            "else": ["$$ids"],
                                        }
                                    },
                                ]
                            }
                        }
                    }
                ]
                
                if ref_sort_dict:
                    lookup_pipeline.append({"$sort": ref_sort_dict})
                if view_fields:
                    lookup_pipeline.append({"$project": {f: 1 for f in view_fields}})

                joined_var = f"_joined_{embed_field}"

                # Step 1: Extract IDs
                if is_nested_local:
                    id_var = f"_ids_{embed_field}"
                    current_pipeline.append(
                        {
                            "$set": {
                                id_var: self._build_dynamic_id_extraction(
                                    path_parts[:-1], local_key
                                )
                            }
                        }
                    )
                else:
                    id_var = local_path

                # Step 2: Lookup
                current_pipeline.append(
                    {
                        "$lookup": {
                            "from": from_collection,
                            "let": {"ids": f"${id_var}"},
                            "pipeline": lookup_pipeline,
                            "as": joined_var,
                        }
                    }
                )

                # Step 3: Embed result
                if is_nested_alias:
                    embed_expr = self._build_dynamic_embedding(
                        alias_parts, joined_var, foreign_field, local_key
                    )
                    current_pipeline.append({"$set": {alias_parts[0]: embed_expr}})
                else:
                    if is_single:
                        current_pipeline.append(
                            {
                                "$set": {
                                    alias: {
                                        "$cond": {
                                            "if": {"$gt": [{"$size": f"${joined_var}"}, 0]},
                                            "then": {"$arrayElemAt": [f"${joined_var}", 0]},
                                            "else": None,
                                        }
                                    }
                                }
                            }
                        )
                    else:
                        current_pipeline.append({"$set": {alias: f"${joined_var}"}})

                if is_nested_local:
                    current_pipeline.append({"$unset": [id_var]})
                current_pipeline.append({"$unset": [joined_var]})

                return current_pipeline

            for ref in references:
                pipeline = process_reference(ref, pipeline)

            return pipeline

    def _build_dynamic_id_extraction(self, path_parts, local_key):
        """Build dynamic ID extraction for any depth with safe $map/$reduce."""
        if len(path_parts) == 1:
            return {
                "$map": {
                    "input": {
                        "$cond": {
                            "if": {"$isArray": f"${path_parts[0]}"},
                            "then": f"${path_parts[0]}",
                            "else": [],
                        }
                    },
                    "as": "item",
                    "in": f"$$item.{local_key}",
                }
            }
        else:
            return {
                "$reduce": {
                    "input": {
                        "$cond": {
                            "if": {"$isArray": f"${path_parts[0]}"},
                            "then": f"${path_parts[0]}",
                            "else": [],
                        }
                    },
                    "initialValue": [],
                    "in": {
                        "$concatArrays": [
                            "$$value",
                            self._extract_nested_ids(
                                "$$this", path_parts[1:], local_key
                            ),
                        ]
                    },
                }
            }

    def _extract_nested_ids(self, parent_expr, remaining_path, local_key):
        """Recursively extract IDs with safe $map/$reduce logic."""
        current_expr = f"{parent_expr}.{remaining_path[0]}"

        if len(remaining_path) == 1:
            return {
                "$cond": {
                    "if": {"$isArray": current_expr},
                    "then": {
                        "$map": {
                            "input": current_expr,
                            "as": "nested_item",
                            "in": f"$$nested_item.{local_key}",
                        }
                    },
                    "else": [],
                }
            }
        else:
            return {
                "$reduce": {
                    "input": {
                        "$cond": {
                            "if": {"$isArray": current_expr},
                            "then": current_expr,
                            "else": [],
                        }
                    },
                    "initialValue": [],
                    "in": {
                        "$concatArrays": [
                            "$$value",
                            self._extract_nested_ids(
                                "$$this", remaining_path[1:], local_key
                            ),
                        ]
                    },
                }
            }

    def _build_dynamic_embedding(
        self,
        alias_parts: List[str],
        joined_var: str,
        foreign_field: str,
        local_key: str,
    ) -> Any:
        """Wrapper to initiate nested embedding from top-level alias"""
        if len(alias_parts) == 1:
            return f"${alias_parts[0]}"  # Just return if there's no nesting

        return self._build_nested_embedding(
            f"${alias_parts[0]}", alias_parts[1:], joined_var, foreign_field, local_key
        )

    def _build_nested_embedding(
        self,
        current_expr: str,
        remaining_alias_parts: List[str],
        joined_var: str,
        foreign_field: str,
        local_key: str,
    ) -> Dict[str, Any]:
        """Recursively build nested embedding expressions with safe map for any depth."""

        # Always guard map input to avoid non-array error
        safe_input = {
            "$cond": {
                "if": {"$isArray": current_expr},
                "then": current_expr,
                "else": [],
            }
        }

        field = remaining_alias_parts[0]
        remaining_path = remaining_alias_parts[1:]

        # Base case
        if not remaining_path:
            return {
                "$map": {
                    "input": safe_input,
                    "as": "parent",
                    "in": {
                        "$mergeObjects": [
                            "$$parent",
                            {
                                field: {
                                    "$cond": {
                                        "if": {"$isArray": f"${joined_var}"},
                                        "then": {
                                            "$filter": {
                                                "input": f"${joined_var}",
                                                "as": "j",
                                                "cond": {
                                                    "$eq": [
                                                        f"$$j.{foreign_field}",
                                                        f"$$parent.{local_key}",
                                                    ]
                                                },
                                            }
                                        },
                                        "else": [],
                                    }
                                }
                            },
                        ]
                    },
                }
            }

        # Recursive case
        return {
            "$map": {
                "input": safe_input,
                "as": "parent",
                "in": {
                    "$mergeObjects": [
                        "$$parent",
                        {
                            field: self._build_nested_embedding(
                                f"$$parent.{field}",
                                remaining_path,
                                joined_var,
                                foreign_field,
                                local_key,
                            )
                        },
                    ]
                },
            }
        }

    def _build_group_stage(self, group_agg: Aggregations, field_filter) -> Dict[str, Any]:
        _id = (
            {field: f"${field}" for field in group_agg.group_by}
            if group_agg.group_by
            else None
        )
        group_stage = {"_id": _id}

 
        group_stage["data"] = {"$push": "$$ROOT"}


        if group_agg.aggregation_fields:
            for agg in group_agg.aggregation_fields:
                agg_type = agg.type.lower()
                field = agg.field
                alias = agg.alias

                if agg_type == "count":
                    group_stage[alias] = {"$sum": 1}
                elif agg_type in AggregationType:
                    group_stage[alias] = {f"${agg_type}": f"${field}"}
                else:
                    raise ValueError(f"Unsupported aggregation type: {agg_type}")
        return {"$group": group_stage}
    
    def _build_aggregation_projection(self, find: ListFilter) -> Dict[str, Any]:
        # Handle field projection before group stage for include type
        if find.field_filter and find.field_filter.fields and find.field_filter.type == "include":
            projection = self._build_field_filters(find.field_filter)
            # Ensure group_by fields are always included
            for field in find.aggregations.group_by:
                projection[field] = 1
            
        elif find.field_filter and find.field_filter.fields and find.field_filter.type == "exclude":
            # Remove group_by fields from exclusion
            for field in find.aggregations.group_by or []:
                if field in find.field_filter.fields:
                    find.field_filter.fields.remove(field)
            projection = self._build_field_filters(find.field_filter)
            
        elif (find.aggregations.document_inclusion_mode == "partial" and 
            find.aggregations.included_fields):
            # Create a temporary field_filter for partial mode
            temp_field_filter = type('obj', (object,), {
                'type': 'include',
                'fields': find.aggregations.included_fields
            })
            projection = self._build_field_filters(temp_field_filter)
            # Ensure group_by fields are always included
            for field in find.aggregations.group_by:
                projection[field] = 1
        else:
            return None
        
        return projection
    


    def _build_group_filter_stage(self, group_agg: Aggregations) -> list:
        try:
            group_options = getattr(group_agg, 'group_options', None)
            if group_options and getattr(group_options, 'group_id', None):
                # Convert values that look like ObjectId to proper ObjectId objects
                match_conditions = {}
                for k, v in group_options.group_id.items():
                    try:
                        if isinstance(v, list):
                            # Handle list of values
                            converted_values = []
                            for item in v:
                                if isinstance(item, str) and ObjectId.is_valid(item):
                                    converted_values.append(ObjectId(item))
                                else:
                                    converted_values.append(item)
                            match_conditions[f"_id.{k}"] = converted_values
                        else:
                            # Handle single value
                            if isinstance(v, str) and ObjectId.is_valid(v):
                                match_conditions[f"_id.{k}"] = ObjectId(v)
                            else:
                                match_conditions[f"_id.{k}"] = v
                    except (TypeError, ValueError):
                        # If conversion fails, use the original value
                        match_conditions[f"_id.{k}"] = v
                
                stage = [{"$match": match_conditions}]
                return stage
            return []
        except Exception as e:
            return []

    def _build_data_pagination_stage(self, group_agg: Aggregations) -> dict:
        try:
            group_options = getattr(group_agg, 'group_options', None)
            data_page = getattr(group_options, 'page', 1) if group_options else 1
            data_page_size = getattr(group_options, 'page_size', 10) if group_options else 10
            data_skip = (data_page - 1) * data_page_size

            sort_by = {"_id": -1}  # Default to descending
            if group_options and getattr(group_options, 'sort_by', None):
                sort_order = group_options.sort_by
                if isinstance(sort_order, list) and len(sort_order) > 0:
                    sort_by = {}
                    for sort_item in sort_order:
                        direction = 1 if sort_item.direction == SortDirection.ASC else -1
                        sort_by[sort_item.field] = direction

            stage = {
                "$set": {
                    "data": {
                        "$slice": [
                            {
                                "$sortArray": {
                                    "input": "$data",
                                    "sortBy": sort_by
                                }
                            },
                            data_skip,
                            data_page_size
                        ]
                    }
                }
            }
            return stage
        except Exception as e:
            # Fallback to default sorting without pagination
            return {
                "$set": {
                    "data": {
                        "$sortArray": {
                            "input": "$data",
                            "sortBy": {"_id": -1}
                        }
                    }
                }
            }

    def _build_count_pipeline(self, pipeline: list) -> list:
        """Build count pipeline for total groups calculation (matches old logic)"""
        count_pipeline = pipeline[:]
        # Remove $set stage for count
        if count_pipeline and "$set" in count_pipeline[-1]:
            count_pipeline = count_pipeline[:-1]
        count_pipeline.append({"$count": "total"})
        return count_pipeline

    def _process_group_data(self, data: list, find: ListFilter) -> tuple:
        try:
            group_options = getattr(find.aggregations, 'group_options', None)
            data_page = getattr(group_options, 'page', 1) if group_options else 1
            data_page_size = getattr(group_options, 'page_size', 10) if group_options else 10
            for group in data:
                try:
                    group['data'] = self._convert_to_json_serializable(group.get("data", []))
                    total_records_in_group = group.get("count", len(group.get("data", [])))
                    total_data_pages = (total_records_in_group + data_page_size - 1) // data_page_size
                    group["data_pagination"] = {
                        "page": data_page,
                        "page_size": data_page_size,
                        "total_pages": total_data_pages,
                        "total_counts": total_records_in_group
                    }
                except Exception as e:
                    group["data_pagination"] = {
                        "page": data_page,
                        "page_size": data_page_size,
                        "total_pages": 0,
                        "total_counts": 0
                    }
            return data, data_page, data_page_size
        except Exception as e:
            return [], 1, 10

    async def delete(
        self, id: Union[str, ObjectId], collection_name: Optional[str] = None, embedding_config: Optional[EmbeddingSchema] = None
    ) -> bool:
        """Delete a document by id"""
        try:
            object_id = ObjectId(id) if isinstance(id, str) else id

            if collection_name:
                collection = self.get_collection(collection_name)
                result = await collection.delete_one({"_id": object_id})
                if result.deleted_count == 0:
                    raise NotFoundError(detail="Document not found")
                if embedding_config and self.azure_service_bus:
                    try:
                        payload = {
                            "event_id": str(uuid4()),
                            "event_key": "embedding",
                            "action": "deletion",
                            "container_name": embedding_config.container_name,
                            "payload": [
                                {
                                    "id": str(object_id),
                                    "text": "",
                                    "metadata": None
                                }
                            ]
                        }
                        await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                    except Exception as e:
                        logger.error(f"Error sending embedding payload: {e}")
                    
                if self.sio_service or self.azure_service_bus:
                    payload = {
                        "event_id": str(uuid4()),
                        "event_key": ON_RECORD_DELETION,
                        "result": result.deleted_count,
                        "payload": None,
                        "record_id": str(id),
                        "collection_name": collection_name,
                    }
                    entity_id = collection_name.split("_")[1]
                    if self.sio_service:
                        await self.sio_service.emit_to_room(
                            ON_RECORD_DELETION,f"entity:{entity_id}" ,payload
                        )
                    if self.azure_service_bus:
                        await self.azure_service_bus.send(payload)
            else:
                result = await self.document.get(object_id)
                if not result:
                    raise NotFoundError(detail="Document not found")
                await result.delete()

            return True

        except Exception as e:
            await self._handle_mongo_error("delete document", e)

    async def get_unique_values(
        self, schema: FindUniqueByFieldInput, collection_name: Optional[str] = None
    ) -> Dict[str, Any]:
        """Get unique values for a field with pagination and search"""
        try:
            if not schema.field_name:
                raise BadRequestError(detail="Field name is required")

            collection = (
                self.get_collection(collection_name)
                if collection_name
                else self.document.get_motor_collection()
            )

            pipeline = []

            if schema.search:
                pipeline.append(
                    {
                        "$match": {
                            schema.field_name: {
                                "$regex": f".*{schema.search}.*",
                                "$options": "i",
                            }
                        }
                    }
                )

            pipeline.append({"$group": {"_id": f"${schema.field_name}"}})

            if schema.ordering:
                pipeline.append(
                    {
                        "$sort": {
                            "_id": ASCENDING if schema.ordering == "asc" else DESCENDING
                        }
                    }
                )

            count_pipeline = pipeline + [{"$count": "total"}]
            count_result = await collection.aggregate(count_pipeline).to_list(length=1)
            total_count = count_result[0]["total"] if count_result else 0

            skip = ((schema.page or 1) - 1) * (schema.page_size or 10)
            pipeline.extend([{"$skip": skip}, {"$limit": schema.page_size or 10}])

            results = await collection.aggregate(pipeline).to_list(length=None)
            values = [r["_id"] for r in results]

            return {
                "founds": values,
                "search_options": {
                    "page": schema.page or 1,
                    "page_size": schema.page_size or 10,
                    "ordering": schema.ordering or "asc",
                    "total_count": total_count,
                },
            }

        except Exception as e:
            await self._handle_mongo_error("get unique values", e)

    async def bulk_update(
        self,
        conditions: FilterSchema,
        update_data: Dict[str, Any],
        collection_name: Optional[str] = None,
        embedding_config: Optional[EmbeddingSchema] = None,
    ) -> int:
        """
        Update multiple documents based on ListFilter conditions

        Args:
            conditions: FilterSchema object containing filter conditions
            update_data: Dictionary of fields to update
            collection_name: Optional collection name

        Returns:
            Dictionary containing update results
        """
        try:
            conditions = conditions.model_dump()
            # Build the MongoDB filter from ListFilter
            mongo_filter = self._build_filter_condition(conditions, collection_name)

            # Coerce the update data to ensure proper type conversion
            update_data = self._coerce_document(update_data)

            # Add updated_at timestamp
            update_data["updated_at"] = datetime.now(UTC)
            entity_id = collection_name.split("_")[1]

            if mongo_filter:
                if collection_name:
                    collection = self.get_collection(collection_name)
                    data = await collection.find(mongo_filter).to_list(length=None)
                    ids = [str(item.get("_id", "")) for item in data]
                    result = await collection.update_many(
                        mongo_filter, {"$set": update_data}
                    )
                    try:
                        if embedding_config and self.azure_service_bus:
                            updated_docs = await collection.find({"_id": {"$in": ids}}).to_list(length=None)
                            payloads = []
                            for doc in updated_docs:
                                doc_data = {}
                                doc_data["id"] = str(doc.get("_id", ""))
                                doc_data["text"] = convert_to_text(doc)
                                doc_data["metadata"] = {
                                    "entity_id": entity_id,
                                    "content":{
                                        "id":str(doc.get("_id", "")),
                                        "created_at":doc.get("created_at", ""),
                                        "updated_at":doc.get("updated_at", ""),
                                        "uuid":doc.get("uuid", ""),
                                        "data":doc
                                    }
                                }
                                payloads.append(self._convert_to_json_serializable(doc_data))
                            payload = {
                                "event_id": str(uuid4()),
                                "event_key": "embedding",
                                "action": "updation",
                                "container_name": embedding_config.container_name,
                                "payload": payloads
                            }
                            await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                    except Exception as e:
                        logger.error(f"Error sending embedding payload: {e}")
                        
                    # TODO: Add event handling for bulk update
                    # payload = {
                    #     "event_id": str(uuid4()),
                    #     "event_key": "records_updated",
                    #     "payload": result,
                    #     "request_body": {
                    #         "conditions": conditions,
                    #         "update_data": update_data,
                    #     },
                    #     "collection_name": collection_name,
                    # }
                    # if self.sio_service:
                    #     await self.sio_service.emit_broadcast(
                    #         "records_updated", payload
                    #     )
                    # if self.azure_service_bus:
                    #     await self.azure_service_bus.send(payload)
                    return {
                        "modified_count": result.modified_count,
                        "matched_count": result.matched_count,
                        "upserted_id": result.upserted_id,
                    }
                else:
                    result = await self.document.find(mongo_filter).update_many(
                        Set(update_data)
                    )
                    return {
                        "modified_count": result.modified_count,
                        "matched_count": result.matched_count,
                        "upserted_id": result.upserted_id,
                    }
            else:
                raise BadRequestError(detail="No filter conditions provided")

        except Exception as e:
            await self._handle_mongo_error("bulk update documents", e)

    async def bulk_delete(
        self, conditions: FilterSchema, collection_name: Optional[str] = None, embedding_config: Optional[EmbeddingSchema] = None
    ) -> int:
        """
        Delete multiple documents based on ListFilter conditions

        Args:
            conditions: FilterSchema object containing filter conditions
            collection_name: Optional collection name

        Returns:
            Dictionary containing deletion results
        """
        try:
            # Build the MongoDB filter from ListFilter
            if hasattr(conditions, "model_dump"):
                conditions = conditions.model_dump(exclude_none=True)
            else:
                conditions = {k: v for k, v in conditions.items() if v is not None}
            mongo_filter = self._build_filter_condition(conditions, collection_name)
            if mongo_filter:
                if collection_name:
                    collection = self.get_collection(collection_name)
                    data = await collection.find(mongo_filter).to_list(length=None)
                    ids = [str(item.get("_id", "")) for item in data]
                    result = await collection.delete_many(mongo_filter)
                    try:
                        if embedding_config and self.azure_service_bus and ids:
                            payloads = []
                            for id in ids:
                                doc_data = {}
                                doc_data["id"] = str(id)
                                payloads.append(self._convert_to_json_serializable(doc_data))
                            payload = {
                                "event_id": str(uuid4()),
                                "event_key": "embedding",
                                "action": "deletion",
                                "container_name": embedding_config.container_name,
                                "payload": payloads
                            }
                            await self.azure_service_bus.send(self._convert_to_json_serializable(payload))
                    except Exception as e:
                        logger.error(f"Error sending embedding payload: {e}")
                    result = await collection.delete_many(mongo_filter)
                    # TODO: Add event handling for bulk delete
                    # payload = {
                    #     "event_id": str(uuid4()),
                    #     "event_key": "records_deleted",
                    #     "payload": result.deleted_count or None,
                    #     "request_body": conditions or None,
                    #     "collection_name": collection_name,
                    # }
                    # if self.sio_service:
                    #     await self.sio_service.emit_broadcast(
                    #         "records_deleted", payload
                    #     )
                    # if self.azure_service_bus:
                    #     await self.azure_service_bus.send(payload)
                    return {"deleted_count": result.deleted_count}
                else:
                    result = await self.document.find(mongo_filter).delete_many()
                    return {"deleted_count": result.deleted_count}
            else:
                raise BadRequestError(detail="No filter conditions provided")

        except Exception as e:
            await self._handle_mongo_error("bulk delete documents", e)


    def _build_field_filters(self, field_filter) -> Dict[str, Any]:
        if not field_filter or not field_filter.fields:
            return {}
        if len(field_filter.fields) == 0:
            return {}

        # Always include _id at the root (or exclude if user wants)
        projection = {"_id": 1} if field_filter.type == "include" else {}
        for field in field_filter.fields:
            if isinstance(field, str) and field != "_id":
                projection[field] = 1 if field_filter.type == "include" else 0
            elif isinstance(field, dict):
                for parent_field, nested_fields in field.items():
                    if isinstance(nested_fields, list):
                        nested_projection = self._build_nested_field_projection(
                            nested_fields, field_filter.type
                        )
                        if nested_projection:
                            projection[parent_field] = nested_projection
        return projection

    def _build_nested_field_projection(self, fields: List[Any], filter_type: str) -> Dict[str, Any]:
        projection = {}
        if filter_type == "include":
            projection["_id"] = 1
        for field in fields:
            if isinstance(field, str) and field != "_id":
                projection[field] = 1 if filter_type == "include" else 0
            elif isinstance(field, dict):
                for parent_field, nested_fields in field.items():
                    if isinstance(nested_fields, list):
                        nested_projection = self._build_nested_field_projection(
                            nested_fields, filter_type
                        )
                        if nested_projection:
                            projection[parent_field] = nested_projection
        return projection



from motor.motor_asyncio import AsyncIOMotorDatabase
from ..schema import AddColumnField, FieldTypeEnum, CreateCollectionSchema
from typing import Optional, Any, List, Dict, TypeVar, Generic
from datetime import datetime, UTC
from uuid import uuid4
from pymongo.errors import PyMongoError, DuplicateKeyError, OperationFailure
from abs_exception_core.exceptions import GenericHttpError, BadRequestError, DuplicatedError,PermissionDeniedError
from pymongo import ASCENDING

T = TypeVar('T')

class BaseCollectionRepository(Generic[T]):
    """
    Base repository class for MongoDB collections with common CRUD operations.
    """

    def __init__(self, db: AsyncIOMotorDatabase):
        """
        Initialize the repository with a MongoDB database connection.
        
        Args:
            db (AsyncIOMotorDatabase): MongoDB database instance
        """
        self.db = db
        self._validate_connection()

    def _validate_connection(self) -> None:
        """Validate the database connection."""
        try:
            # Try to get server info to validate connection
            self.db.client.server_info()
        except Exception as e:
            raise ConnectionError(f"Failed to connect to MongoDB: {str(e)}")

    def get_base_document_fields(self) -> Dict[str, Any]:
        """
        Get the base document fields that should be present in all documents.
        These fields match the BaseDocument class fields.
        
        Returns:
            Dict[str, Any]: Dictionary containing base document fields
        """
        return {
            "uuid": str(uuid4()),  # Unique identifier
            "created_at": datetime.now(UTC),  # Creation timestamp
            "updated_at": datetime.now(UTC)  # Last update timestamp
        }

    async def _handle_mongo_error(self, operation: str, error: Exception) -> None:
        """
        Handle MongoDB errors consistently with appropriate error types.
        
        Args:
            operation (str): Name of the operation that failed
            error (Exception): The error that occurred
            
        Raises:
            GenericHttpError: For general MongoDB errors
            BadRequestError: For invalid operations
            NotFoundError: For missing resources
        """
        
        if isinstance(error, DuplicateKeyError):
            raise BadRequestError(
                detail=f"Failed to {operation}: Duplicate key found"
            )
        elif isinstance(error, OperationFailure):
            raise BadRequestError(
                detail=f"Failed to {operation}: {str(error)}"
            )
        elif isinstance(error, PyMongoError):
            raise GenericHttpError(
                status_code=500,
                detail=str(error),
                error_type="PyMongoError",
                message=f"Failed to {operation}"
            )
        raise BadRequestError(detail=str(error))

    async def add_field(self, column: AddColumnField, collection_name: Optional[str] = None) -> bool:
        """
        Add a new field to all documents in the collection.
        
        Args:
            column (AddColumnField): Column configuration
            collection_name (Optional[str]): Name of the collection
            
        Returns:
            bool: True if operation was successful
            
        Raises:
            BadRequestError: If collection name is invalid or operation fails
        """
        if not collection_name:
            raise BadRequestError(detail="Collection name is required")
        
        try:
            collection = self.db[collection_name]
            result = await collection.update_many(
                {column.column_field: {"$exists": False}},
                {"$set": {column.column_field: column.column_default or self.get_default_for_type(column.column_type)}}
            )
            if column.column_index:
                await collection.create_index(column.column_field)
            return True
                    
        except Exception as e:
            await self._handle_mongo_error("add field", e)

    async def delete_field(self, column_name: str, collection_name: Optional[str] ,protected_fields:Optional[List[str]]=None) -> bool:
        """
        Delete a field from all documents in the collection.
        
        Args:
            column_name (str): Name of the field to delete
            collection_name (Optional[str]): Name of the collection
            protected_fields (List[str]): List of protected fields
            
        Returns:
            bool: True if operation was successful
            
        Raises:
            BadRequestError: If collection name is invalid
        """
        if not collection_name:
            raise BadRequestError(detail="Collection name is required")
        if protected_fields is None:
            protected_fields = ["_id","uuid","created_at","updated_at","name"]
        else:
            protected_fields = protected_fields+["_id","uuid","created_at","updated_at","name"]

        if protected_fields and column_name in protected_fields:
            raise PermissionDeniedError("Requested field cannot be deleted.")
            
        try:
            collection = self.db[collection_name]
            result = await collection.update_many(
                {},
                {"$unset": {column_name: ""}}
            )
            
            return True
                    
        except Exception as e:
            await self._handle_mongo_error("delete field", e)

    async def rename_field(self, old_name: str, new_name: str, collection_name: Optional[str] = None) -> bool:
        """
        Rename a field in all documents in the collection.
        
        Args:
            old_name (str): Current name of the field
            new_name (str): New name of the field
            collection_name (Optional[str]): Name of the collection
            
        Returns:
            bool: True if operation was successful
            
        Raises:
            BadRequestError: If collection name is invalid or target field already exists
        """
        if not collection_name:
            raise BadRequestError(detail="Collection name is required")
        
        try:
            collection = self.db[collection_name]
            
            # Check if target field exists in any records
            existing_docs = await collection.find({new_name: {"$exists": True}}).to_list(length=1)
            if existing_docs:
                raise DuplicatedError(
                    detail=f"Cannot rename field to '{new_name}' as it already exists in some records"
                )
            
            result = await collection.update_many(
                {},
                {"$rename": {old_name: new_name}}
            )
            
            return True
                    
        except Exception as e:
            await self._handle_mongo_error("rename field", e)
            
    

    async def create_collection(self, collection_data: CreateCollectionSchema) -> bool:
        """
        Create a new collection with default values.
        
        Args:
            collection_data (CreateCollectionSchema): Collection configuration
            
        Returns:
            bool: True if operation was successful
            
        Raises:
            BadRequestError: If collection name is invalid or operation fails
        """
        
        try:
            collection = self.db[collection_data.collection_name]
            default_fields = collection_data.default_values or {}
            base_fields = self.get_base_document_fields()
            full_doc = {**base_fields, **default_fields}

            collection.create_index([("_id", ASCENDING)])
            collection.create_index([("uuid", ASCENDING)])

            
            await collection.insert_one(full_doc)
            return True
                    
        except Exception as e:
            await self._handle_mongo_error("create collection", e)

    async def delete_collection(self, collection_name: str) -> bool:
        """
        Delete a collection.
        
        Args:
            collection_name (str): Name of the collection to delete
            
        Returns:
            bool: True if operation was successful
        """
        
        try:
            await self.db.drop_collection(collection_name)
            return True
            
        except Exception as e:
            await self._handle_mongo_error("delete collection", e)

    async def rename_collection(self, old_name: str, new_name: str) -> bool:
        """
        Rename a collection in MongoDB.
        
        Args:
            old_name (str): Current collection name
            new_name (str): New collection name
            
        Returns:
            bool: True if operation was successful
            
        Raises:
            BadRequestError: If collection names are invalid
        """
        
        try:
            full_old = f"{self.db.name}.{old_name}"
            full_new = f"{self.db.name}.{new_name}"

            admin_db = self.db.client.admin

            await admin_db.command({
                "renameCollection": full_old,
                "to": full_new,
                "dropTarget": False 
            })

            return True
            
        except Exception as e:
            await self._handle_mongo_error("rename collection", e)

    async def get_collection_names(self) -> List[str]:
        """
        Get all collection names in the database.
        
        Returns:
            List[str]: List of collection names
        """
        try:
            return await self.db.list_collection_names()
        except Exception as e:
            await self._handle_mongo_error("get collection names", e)

    @staticmethod
    def get_default_for_type(field_type: FieldTypeEnum) -> Any:
        """
        Get default value for a field type.
        
        Args:
            field_type (FieldTypeEnum): Type of the field
            
        Returns:
            Any: Default value for the field type
        """
        default_values = {
            FieldTypeEnum.STR: "",
            FieldTypeEnum.INT: 0,
            FieldTypeEnum.BOOL: False,
            FieldTypeEnum.FLOAT: 0.0,
            FieldTypeEnum.LIST: [],
            FieldTypeEnum.DICT: {},
            FieldTypeEnum.DATETIME: datetime.now(UTC),
            FieldTypeEnum.DATE : datetime.now(UTC)
        }
        return default_values.get(field_type) or ""



from typing import List

from ..repository.base_collection_repository import BaseCollectionRepository
from ..schema import CreateCollectionSchema,AddColumnField

class BaseCollectionService:
    def __init__(self, repository: BaseCollectionRepository):
        self.repository = repository

    async def create_collection(self, collection: CreateCollectionSchema):
        return await self.repository.create_collection(collection)
    
    async def delete_collection(self, collection_name: str):
        return await self.repository.delete_collection(collection_name)
    
    async def get_collection_names(self):
        return await self.repository.get_collection_names()
    
    async def rename_collection(self, old_name: str, new_name: str):
        return await self.repository.rename_collection(old_name, new_name)
    
    async def add_field(self, field: AddColumnField, collection_name: str):
        return await self.repository.add_field( field,collection_name)
    
    async def delete_field(self, field: str, collection_name: str,protected_fields:List[str]):
        return await self.repository.delete_field(field,collection_name,protected_fields)
    
    async def rename_field(self, old_name: str, new_name: str, collection_name: str):
        return await self.repository.rename_field(old_name, new_name, collection_name)
    


from pydantic import BaseModel
from typing import Optional, Any, Dict
from datetime import UTC
from enum import Enum

class FieldTypeEnum(str, Enum):
    STR = "str"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    LIST = "list"
    DICT = "dict"
    DATETIME = "datetime"
    DATE = "date"

# Add column field schema
class AddColumnField(BaseModel):
    column_field:str
    column_type:FieldTypeEnum
    column_default:Optional[Any] = None
    column_index:Optional[bool] = False

class CreateCollectionSchema(BaseModel):
    collection_name:str
    default_values:Optional[Dict[str,Any]] = None